@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 
-/* Copyright (C) 2018 Hans Petter Jansson
+/* Copyright (C) 2018-2020 Hans Petter Jansson
  *
  * This file is part of Chafa, a program that turns images into character art.
  *
@@ -37,10 +37,14 @@
 #endif
 
 #include <chafa.h>
+#include "font-loader.h"
 #include "gif-loader.h"
 #include "named-colors.h"
 #include "xwd-loader.h"
 
+#define ANIM_FPS_MAX 100000.0
+#define FILE_DURATION_DEFAULT 0.0
+
 typedef struct
 {
     gchar *executable_name;
@@ -50,8 +54,10 @@ typedef struct
 
     GList *args;
     ChafaCanvasMode mode;
+    ChafaColorExtractor color_extractor;
     ChafaColorSpace color_space;
     ChafaDitherMode dither_mode;
+    ChafaPixelMode pixel_mode;
     gint dither_grain_width;
     gint dither_grain_height;
     gdouble dither_intensity;
@@ -67,6 +73,7 @@ typedef struct
     gboolean zoom;
     gboolean watch;
     gint width, height;
+    gint cell_width, cell_height;
     gdouble font_ratio;
     gint work_factor;
     guint32 fg_color;
@@ -75,6 +82,13 @@ typedef struct
     gboolean bg_color_set;
     gdouble transparency_threshold;
     gdouble file_duration_s;
+
+    /* If > 0.0, override the framerate specified by the input file. */
+    gdouble anim_fps;
+
+    /* Applied after FPS is determined. If final value >= ANIM_FPS_MAX,
+     * eliminate interframe delay altogether. */
+    gdouble anim_speed_multiplier;
 }
 GlobalOptions;
 
@@ -172,7 +186,7 @@ collect_variable_arguments (int *argc, char **argv [], gint first_arg)
 }
 
 static const gchar copyright_notice [] =
-    "Copyright (C) 2018 Hans Petter Jansson et al.\n"
+    "Copyright (C) 2018-2020 Hans Petter Jansson et al.\n"
     "Incl. libnsgif copyright (C) 2004 Richard Wilson, copyright (C) 2008 Sean Fox\n\n"
     "This is free software; see the source for copying conditions.  There is NO\n"
     "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n";
@@ -209,6 +223,8 @@ print_summary (void)
     "      --clear        Clear screen before processing each file.\n"
     "  -c, --colors=MODE  Set output color mode; one of [none, 2, 16, 240, 256,\n"
     "                     full]. Defaults to full (24-bit).\n"
+    "      --color-extractor=EXTR  Method for extracting color from an area\n"
+    "                     [average, median]. Median is the default.\n"
     "      --color-space=CS  Color space used for quantization; one of [rgb, din99d].\n"
     "                     Defaults to rgb, which is faster but less accurate.\n"
     "      --dither=DITHER  Set output dither mode; one of [none, ordered,\n"
@@ -226,6 +242,10 @@ print_summary (void)
     "                     Defaults to none. See below for full usage.\n"
     "      --font-ratio=W/H  Target font's width/height ratio. Can be specified as\n"
     "                     a real number or a fraction. Defaults to 1/2.\n"
+    "  -f, --format=FORMAT  Set output format; one of [sixels, symbols]. Sixels\n"
+    "                     yield much better quality but are rarely supported.\n"
+    "      --glyph-file=FILE  Load glyph information from FILE, which can be any\n"
+    "                     font file supported by FreeType (TTF, PCF, etc).\n"
     "      --invert       Invert video. For display with bright backgrounds in\n"
     "                     color modes 2 and none. Swaps --fg and --bg.\n"
     "  -p, --preprocess=BOOL  Image preprocessing [on, off]. Defaults to on with 16\n"
@@ -233,6 +253,9 @@ print_summary (void)
     "  -s, --size=WxH     Set maximum output dimensions in columns and rows. By\n"
     "                     default this will be the size of your terminal, or 80x25\n"
     "                     if size detection fails.\n"
+    "      --speed=SPEED  Set the speed animations will play at. This can be\n"
+    "                     either a unitless multiplier, or a real number followed\n"
+    "                     by \"fps\" to apply a specific framerate.\n"
     "      --stretch      Stretch image to fit output dimensions; ignore aspect.\n"
     "                     Implies --zoom.\n"
     "      --symbols=SYMS  Specify character symbols to employ in final output.\n"
@@ -301,6 +324,25 @@ parse_colors_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_
     return result;
 }
 
+static gboolean
+parse_color_extractor_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+{
+    gboolean result = TRUE;
+
+    if (!g_ascii_strcasecmp (value, "average"))
+        options.color_extractor = CHAFA_COLOR_EXTRACTOR_AVERAGE;
+    else if (!g_ascii_strcasecmp (value, "median"))
+        options.color_extractor = CHAFA_COLOR_EXTRACTOR_MEDIAN;
+    else
+    {
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Color extractor must be one of [average, median].");
+        result = FALSE;
+    }
+
+    return result;
+}
+
 static gboolean
 parse_color_space_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
 {
@@ -396,6 +438,35 @@ parse_fill_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GN
     return chafa_symbol_map_apply_selectors (options.fill_symbol_map, value, error);
 }
 
+static gboolean
+parse_format_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+{
+    gboolean result = TRUE;
+    ChafaPixelMode pixel_mode = CHAFA_PIXEL_MODE_SYMBOLS;
+
+    if (!strcasecmp (value, "symbol") || !strcasecmp (value, "symbols")
+        || !strcasecmp (value, "ansi"))
+    {
+        pixel_mode = CHAFA_PIXEL_MODE_SYMBOLS;
+    }
+    else if (!strcasecmp (value, "sixel") || !strcasecmp (value, "sixels"))
+    {
+        pixel_mode = CHAFA_PIXEL_MODE_SIXELS;
+    }
+    else
+    {
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Output format given as '%s'. Must be one of [symbols, sixels].",
+                     value);
+        result = FALSE;
+    }
+
+    if (result)
+        options.pixel_mode = pixel_mode;
+
+    return result;
+}
+
 static gboolean
 parse_size_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
 {
@@ -484,6 +555,56 @@ out:
     return result;
 }
 
+static gboolean
+parse_glyph_file_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+{
+    gboolean result = FALSE;
+    FileMapping *file_mapping;
+    FontLoader *font_loader;
+    gunichar c;
+    gpointer c_bitmap;
+    gint width, height;
+
+    file_mapping = file_mapping_new (value);
+    if (!file_mapping)
+    {
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Unable to open glyph file '%s'.", value);
+        goto out;
+    }
+
+    font_loader = font_loader_new_from_mapping (file_mapping);
+    if (!font_loader)
+    {
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Unable to load glyph file '%s'.", value);
+        goto out;
+    }
+
+    /* Font loader owns it now */
+    file_mapping = NULL;
+
+    while (font_loader_get_next_glyph (font_loader, &c, &c_bitmap, &width, &height))
+    {
+        chafa_symbol_map_add_glyph (options.symbol_map, c,
+                                    CHAFA_PIXEL_RGBA8_PREMULTIPLIED, c_bitmap,
+                                    width, height, width * 4);
+        chafa_symbol_map_add_glyph (options.fill_symbol_map, c,
+                                    CHAFA_PIXEL_RGBA8_PREMULTIPLIED, c_bitmap,
+                                    width, height, width * 4);
+        g_free (c_bitmap);
+    }
+
+    font_loader_destroy (font_loader);
+
+    result = TRUE;
+
+out:
+    if (file_mapping)
+        file_mapping_destroy (file_mapping);
+    return result;
+}
+
 static gboolean
 parse_preprocess_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
 {
@@ -509,6 +630,51 @@ parse_preprocess_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value
     return result;
 }
 
+static gboolean
+parse_anim_speed_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+{
+    gboolean result = FALSE;
+    gdouble d;
+
+    if (!g_ascii_strcasecmp (value, "max")
+        || !g_ascii_strcasecmp (value, "maximum"))
+    {
+        options.anim_fps = G_MAXDOUBLE;
+        result = TRUE;
+    }
+    else
+    {
+        gchar *endptr;
+
+        d = g_strtod (value, &endptr);
+        if (endptr == value || d <= 0.0)
+            goto out;
+
+        while (g_ascii_isspace (*endptr))
+            endptr++;
+
+        if (!g_ascii_strcasecmp (endptr, "fps"))
+        {
+            options.anim_fps = d;
+            result = TRUE;
+        }
+        else if (*endptr == '\0')
+        {
+            options.anim_speed_multiplier = d;
+            result = TRUE;
+        }
+
+        /* If there's unknown junk at the end, fail. */
+    }
+
+out:
+    if (!result)
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Animation speed must be either \"max\", a real multiplier, or a real framerate followed by \"fps\". It must be greater than zero.");
+
+    return result;
+}
+
 static gboolean
 parse_color_str (const gchar *value, guint32 *col_out, const gchar *error_message, GError **error)
 {
@@ -565,10 +731,18 @@ get_tty_size (void)
     if (w.ws_row > 2)
         options.height = w.ws_row - 1;
 
-    /* If .ws_xpixel and .ws_ypixel were filled out, we could in theory
-     * calculate aspect information for the font used. Unfortunately,
-     * I have yet to find an environment where these fields get set to
-     * anything other than zero. */
+    /* If .ws_xpixel and .ws_ypixel are filled out, we can calculate
+     * aspect information for the font used. Sixel-capable terminals
+     * like mlterm set these fields, but most others do not. */
+
+    if (w.ws_xpixel > 0 && w.ws_ypixel > 0)
+    {
+        options.cell_width = w.ws_xpixel / w.ws_col;
+        options.cell_height = w.ws_ypixel / w.ws_row;
+
+        if (options.cell_width > 0 && options.cell_height > 0)
+            options.font_ratio = (gfloat) options.cell_width / (gfloat) options.cell_height;
+    }
 }
 
 static struct termios saved_termios;
@@ -602,14 +776,15 @@ tty_options_deinit (void)
  *
  * If you're getting poor results, I'd love to hear about it so it can
  * be improved. */
-static ChafaCanvasMode
-detect_canvas_mode (void)
+static void
+detect_terminal_modes (ChafaCanvasMode *mode_out, ChafaPixelMode *pixel_mode_out)
 {
     const gchar *term;
     const gchar *colorterm;
     const gchar *vte_version;
     const gchar *tmux;
     ChafaCanvasMode mode = CHAFA_CANVAS_MODE_INDEXED_240;
+    ChafaPixelMode pixel_mode = CHAFA_PIXEL_MODE_SYMBOLS;
 
     term = g_getenv ("TERM");
     if (!term) term = "";
@@ -644,9 +819,19 @@ detect_canvas_mode (void)
         mode = CHAFA_CANVAS_MODE_TRUECOLOR;
 
     /* mlterm's truecolor support seems to be broken; it looks like a color
-     * allocation issue. */
-    if (!strcmp (term, "mlterm"))
+     * allocation issue. This affects character cells, but not sixels.
+     *
+     * yaft supports sixels and truecolor escape codes, but it remaps cell
+     * colors to a 256-color palette. */
+    if (!strcmp (term, "mlterm")
+        || !strcmp (term, "yaft")
+        || !strcmp (term, "yaft-256color"))
+    {
+        /* The default canvas mode is truecolor for sixels. 240 colors is
+         * the default for symbols. */
         mode = CHAFA_CANVAS_MODE_INDEXED_240;
+        pixel_mode = CHAFA_PIXEL_MODE_SIXELS;
+    }
 
     /* rxvt 256-color really is 256 colors only */
     if (!strcmp (term, "rxvt-unicode-256color"))
@@ -683,7 +868,8 @@ detect_canvas_mode (void)
     if (!strcmp (term, "linux"))
         mode = CHAFA_CANVAS_MODE_INDEXED_16;
 
-    return mode;
+    *mode_out = mode;
+    *pixel_mode_out = pixel_mode;
 }
 
 static gboolean
@@ -702,6 +888,7 @@ parse_options (int *argc, char **argv [])
         { "bg",          '\0', 0, G_OPTION_ARG_CALLBACK, parse_bg_color_arg,    "Background color of display", NULL },
         { "clear",       '\0', 0, G_OPTION_ARG_NONE,     &options.clear,        "Clear", NULL },
         { "colors",      'c',  0, G_OPTION_ARG_CALLBACK, parse_colors_arg,      "Colors (none, 2, 16, 256, 240 or full)", NULL },
+        { "color-extractor", '\0', 0, G_OPTION_ARG_CALLBACK, parse_color_extractor_arg, "Color extractor (average or median)", NULL },
         { "color-space", '\0', 0, G_OPTION_ARG_CALLBACK, parse_color_space_arg, "Color space (rgb or din99d)", NULL },
         { "dither",      '\0', 0, G_OPTION_ARG_CALLBACK, parse_dither_arg,      "Dither", NULL },
         { "dither-grain",'\0', 0, G_OPTION_ARG_CALLBACK, parse_dither_grain_arg, "Dither grain", NULL },
@@ -709,11 +896,14 @@ parse_options (int *argc, char **argv [])
         { "duration",    'd',  0, G_OPTION_ARG_DOUBLE,   &options.file_duration_s, "Duration", NULL },
         { "fg",          '\0', 0, G_OPTION_ARG_CALLBACK, parse_fg_color_arg,    "Foreground color of display", NULL },
         { "fill",        '\0', 0, G_OPTION_ARG_CALLBACK, parse_fill_arg,        "Fill symbols", NULL },
+        { "format",      'f',  0, G_OPTION_ARG_CALLBACK, parse_format_arg,      "Format of output pixel data (ansi or sixels)", NULL },
         { "font-ratio",  '\0', 0, G_OPTION_ARG_CALLBACK, parse_font_ratio_arg,  "Font ratio", NULL },
+        { "glyph-file",  '\0', 0, G_OPTION_ARG_CALLBACK, parse_glyph_file_arg,  "Glyph file", NULL },
         { "invert",      '\0', 0, G_OPTION_ARG_NONE,     &options.invert,       "Invert foreground/background", NULL },
         { "preprocess",  'p',  0, G_OPTION_ARG_CALLBACK, parse_preprocess_arg,  "Preprocessing", NULL },
         { "work",        'w',  0, G_OPTION_ARG_INT,      &options.work_factor,  "Work factor", NULL },
         { "size",        's',  0, G_OPTION_ARG_CALLBACK, parse_size_arg,        "Output size", NULL },
+        { "speed",       '\0', 0, G_OPTION_ARG_CALLBACK, parse_anim_speed_arg,  "Animation speed", NULL },
         { "stretch",     '\0', 0, G_OPTION_ARG_NONE,     &options.stretch,      "Stretch image to fix output dimensions", NULL },
         { "symbols",     '\0', 0, G_OPTION_ARG_CALLBACK, parse_symbols_arg,     "Output symbols", NULL },
         { "threshold",   't',  0, G_OPTION_ARG_DOUBLE,   &options.transparency_threshold, "Transparency threshold", NULL },
@@ -721,6 +911,8 @@ parse_options (int *argc, char **argv [])
         { "zoom",        '\0', 0, G_OPTION_ARG_NONE,     &options.zoom,         "Allow scaling up beyond one character per pixel", NULL },
         { NULL }
     };
+    ChafaCanvasMode canvas_mode;
+    ChafaPixelMode pixel_mode;
 
     memset (&options, 0, sizeof (options));
 
@@ -743,21 +935,26 @@ parse_options (int *argc, char **argv [])
     options.fill_symbol_map = chafa_symbol_map_new ();
 
     options.is_interactive = isatty (STDIN_FILENO) && isatty (STDOUT_FILENO);
-    options.mode = detect_canvas_mode ();
+    detect_terminal_modes (&canvas_mode, &pixel_mode);
+    options.mode = CHAFA_CANVAS_MODE_MAX;  /* Unset */
+    options.pixel_mode = pixel_mode;
     options.dither_mode = CHAFA_DITHER_MODE_NONE;
-    options.dither_grain_width = 4;
-    options.dither_grain_height = 4;
+    options.dither_grain_width = -1;  /* Unset */
+    options.dither_grain_height = -1;  /* Unset */
     options.dither_intensity = 1.0;
     options.preprocess = TRUE;
+    options.color_extractor = CHAFA_COLOR_EXTRACTOR_MEDIAN;
     options.color_space = CHAFA_COLOR_SPACE_RGB;
     options.width = 80;
     options.height = 25;
-    options.font_ratio = 1.0 / 2.0;
+    options.font_ratio = -1.0;  /* Unset */
     options.work_factor = 5;
     options.fg_color = 0xffffff;
     options.bg_color = 0x000000;
     options.transparency_threshold = -1.0;
     options.file_duration_s = G_MAXDOUBLE;
+    options.anim_fps = -1.0;
+    options.anim_speed_multiplier = 1.0;
     get_tty_size ();
 
     if (!g_option_context_parse (context, argc, argv, &error))
@@ -766,6 +963,35 @@ parse_options (int *argc, char **argv [])
         return FALSE;
     }
 
+    /* Now we've established the pixel mode, apply dependent defaults */
+
+    if (options.pixel_mode == CHAFA_PIXEL_MODE_SYMBOLS)
+    {
+        /* Character cell defaults */
+
+        if (options.mode == CHAFA_CANVAS_MODE_MAX)
+            options.mode = canvas_mode;
+        if (options.dither_grain_width < 0)
+            options.dither_grain_width = 4;
+        if (options.dither_grain_height < 0)
+            options.dither_grain_height = 4;
+        if (options.font_ratio < 0.0)
+            options.font_ratio = 1.0 / 2.0;
+    }
+    else
+    {
+        /* Sixel defaults */
+
+        if (options.mode == CHAFA_CANVAS_MODE_MAX)
+            options.mode = CHAFA_CANVAS_MODE_TRUECOLOR;
+        if (options.dither_grain_width < 0)
+            options.dither_grain_width = 1;
+        if (options.dither_grain_height < 0)
+            options.dither_grain_height = 1;
+        if (options.font_ratio < 0.0)
+            options.font_ratio = 1.0;
+    }
+
     if (options.work_factor < 1 || options.work_factor > 9)
     {
         g_printerr ("%s: Work factor must be in the range [1-9].\n", options.executable_name);
@@ -813,7 +1039,7 @@ parse_options (int *argc, char **argv [])
     if (options.file_duration_s == G_MAXDOUBLE && options.args && options.args->next)
     {
         /* The default duration when we have multiple files */
-        options.file_duration_s = 3.0;
+        options.file_duration_s = FILE_DURATION_DEFAULT;
     }
 
     /* Since FGBG mode can't use escape sequences to invert, it really
@@ -833,7 +1059,7 @@ auto_orient_image (MagickWand *image)
 #ifdef HAVE_MAGICK_AUTO_ORIENT_IMAGE
     MagickAutoOrientImage (image);
 #else
-    PixelWand *pwand = 0;
+    PixelWand *pwand = NULL;
 
     switch (MagickGetImageOrientation (image))
     {
@@ -847,6 +1073,7 @@ auto_orient_image (MagickWand *image)
         case BottomRightOrientation:
             pwand = NewPixelWand ();
             MagickRotateImage (image, pwand, 180.0);
+            break;
         case BottomLeftOrientation:
             MagickFlipImage (image);
             break;
@@ -878,25 +1105,27 @@ build_string (ChafaPixelType pixel_type, const guint8 *pixels,
               gint src_width, gint src_height, gint src_rowstride,
               gint dest_width, gint dest_height)
 {
-    ChafaSymbolMap *symbol_map;
     ChafaCanvasConfig *config;
     ChafaCanvas *canvas;
     GString *gs;
 
-    symbol_map = chafa_symbol_map_new ();
     config = chafa_canvas_config_new ();
 
     chafa_canvas_config_set_geometry (config, dest_width, dest_height);
     chafa_canvas_config_set_canvas_mode (config, options.mode);
+    chafa_canvas_config_set_pixel_mode (config, options.pixel_mode);
     chafa_canvas_config_set_dither_mode (config, options.dither_mode);
     chafa_canvas_config_set_dither_grain_size (config, options.dither_grain_width, options.dither_grain_height);
     chafa_canvas_config_set_dither_intensity (config, options.dither_intensity);
+    chafa_canvas_config_set_color_extractor (config, options.color_extractor);
     chafa_canvas_config_set_color_space (config, options.color_space);
     chafa_canvas_config_set_fg_color (config, options.fg_color);
     chafa_canvas_config_set_bg_color (config, options.bg_color);
     chafa_canvas_config_set_preprocessing_enabled (config, options.preprocess);
     if (options.transparency_threshold >= 0.0)
         chafa_canvas_config_set_transparency_threshold (config, options.transparency_threshold);
+    if (options.cell_width > 0 && options.cell_height > 0)
+        chafa_canvas_config_set_cell_geometry (config, options.cell_width, options.cell_height);
 
     chafa_canvas_config_set_symbol_map (config, options.symbol_map);
     chafa_canvas_config_set_fill_symbol_map (config, options.fill_symbol_map);
@@ -911,7 +1140,6 @@ build_string (ChafaPixelType pixel_type, const guint8 *pixels,
 
     chafa_canvas_unref (canvas);
     chafa_canvas_config_unref (config);
-    chafa_symbol_map_unref (symbol_map);
     return gs;
 }
 
@@ -1158,7 +1386,7 @@ run_magickwand (const gchar *filename, gboolean is_first_file, gboolean is_first
              l = g_list_next (l))
         {
             GroupFrame *frame = l->data;
-            gint elapsed_ms, remain_ms;
+            gdouble elapsed_ms, remain_ms;
 
             if (wand)
                 MagickNextImage (wand);
@@ -1241,16 +1469,27 @@ run_magickwand (const gchar *filename, gboolean is_first_file, gboolean is_first
                 fputc ('\n', stdout);
 
             fwrite (frame->gs->str, sizeof (gchar), frame->gs->len, stdout);
-            fputc ('\n', stdout);
+
+            /* No linefeed after frame in sixel mode */
+            if (options.pixel_mode == CHAFA_PIXEL_MODE_SYMBOLS)
+                fputc ('\n', stdout);
+
             fflush (stdout);
 
             if (is_animation)
             {
                 /* Account for time spent converting and printing frame */
                 elapsed_ms = g_timer_elapsed (timer, NULL) * 1000.0;
-                remain_ms = frame->delay_ms - elapsed_ms;
-                remain_ms = MAX (remain_ms, 0);
-                interruptible_usleep (remain_ms * 1000);
+
+                if (options.anim_fps > 0.0)
+                    remain_ms = 1000.0 / options.anim_fps;
+                else
+                    remain_ms = frame->delay_ms;
+                remain_ms /= options.anim_speed_multiplier;
+                remain_ms = MAX (remain_ms - elapsed_ms, 0);
+
+                if (remain_ms > 0.0001 && 1000.0 / (gdouble) remain_ms < ANIM_FPS_MAX)
+                    interruptible_usleep (remain_ms * 1000);
 
                 anim_elapsed_s += MAX (elapsed_ms, frame->delay_ms) / 1000.0;
             }
@@ -1321,7 +1560,7 @@ run_gif (const gchar *filename, gboolean is_first_file, gboolean is_first_frame,
              l = g_list_next (l))
         {
             GroupFrame *frame = l->data;
-            gint elapsed_ms, remain_ms;
+            gdouble elapsed_ms, remain_ms;
 
             g_timer_start (timer);
 
@@ -1377,16 +1616,27 @@ run_gif (const gchar *filename, gboolean is_first_file, gboolean is_first_frame,
                 fputc ('\n', stdout);
 
             fwrite (frame->gs->str, sizeof (gchar), frame->gs->len, stdout);
-            fputc ('\n', stdout);
+
+            /* No linefeed after frame in sixel mode */
+            if (options.pixel_mode == CHAFA_PIXEL_MODE_SYMBOLS)
+                fputc ('\n', stdout);
+
             fflush (stdout);
 
             if (is_animation)
             {
                 /* Account for time spent converting and printing frame */
                 elapsed_ms = g_timer_elapsed (timer, NULL) * 1000.0;
-                remain_ms = frame->delay_ms - elapsed_ms;
-                remain_ms = MAX (remain_ms, 0);
-                interruptible_usleep (remain_ms * 1000);
+
+                if (options.anim_fps > 0.0)
+                    remain_ms = 1000.0 / options.anim_fps;
+                else
+                    remain_ms = frame->delay_ms;
+                remain_ms /= options.anim_speed_multiplier;
+                remain_ms = MAX (remain_ms - elapsed_ms, 0);
+
+                if (remain_ms > 0.0001 && 1000.0 / (gdouble) remain_ms < ANIM_FPS_MAX)
+                    interruptible_usleep (remain_ms * 1000);
 
                 anim_elapsed_s += MAX (elapsed_ms, frame->delay_ms) / 1000.0;
             }
